@@ -6,6 +6,7 @@
 
 module Codec.CBOR.Cuddle.CBOR.Validator (
   validateCBOR,
+  isCBORTermResultValid,
   CDDLResult (..),
   CBORTermResult (..),
   ValidatorStage,
@@ -28,7 +29,6 @@ import Data.Maybe
 import Data.Text qualified as T
 import Data.Text.Lazy qualified as TL
 import Data.Word
-import Debug.Trace (trace, traceShow)
 import GHC.Float
 import GHC.Stack (HasCallStack)
 import Text.Regex.TDFA
@@ -36,10 +36,10 @@ import Text.Regex.TDFA
 type data ValidatorStage
 
 data instance XTerm ValidatorStage = ValidatorXTerm
-  deriving (Show)
+  deriving (Show, Eq)
 
 newtype instance XXCTree ValidatorStage = VRuleRef Name
-  deriving (Show)
+  deriving (Show, Eq)
 
 instance IndexMappable CTreeRoot MonoReferenced ValidatorStage where
   mapIndex (CTreeRoot m) = CTreeRoot $ mapIndex <$> m
@@ -57,7 +57,7 @@ data CBORTermResult = CBORTermResult
   { ctrTerm :: Term
   , ctrResult :: CDDLResult
   }
-  deriving (Show)
+  deriving (Show, Eq)
 
 data CDDLResult
   = -- | The rule was valid
@@ -111,7 +111,11 @@ data CDDLResult
       String
       -- | Rule we are trying
       Rule
-  deriving (Show)
+  deriving (Show, Eq)
+
+isCBORTermResultValid :: CBORTermResult -> Bool
+isCBORTermResultValid (CBORTermResult _ Valid {}) = True
+isCBORTermResultValid _ = False
 
 data ANonMatchedItem = ANonMatchedItem
   { anmiKey :: Term
@@ -120,14 +124,14 @@ data ANonMatchedItem = ANonMatchedItem
   -- ^ For all the tried rules, either the key failed or the key succeeded and
   -- the value failed
   }
-  deriving (Show)
+  deriving (Show, Eq)
 
 data AMatchedItem = AMatchedItem
   { amiKey :: Term
   , amiValue :: Term
   , amiRule :: Rule
   }
-  deriving (Show)
+  deriving (Show, Eq)
 
 --------------------------------------------------------------------------------
 -- Main entry point
@@ -683,13 +687,13 @@ validateList cddl terms rule =
         OIOneOrMore -> case validateTermInList (t : ts) ct of
           (Valid {}, leftover) -> validate leftover (Occur ct OIZeroOrMore : rs)
           (err, _) -> err
-        OIBounded _ (Just ub) | ub < 0 -> trace ("out of bounds: " <> show ub) $ ListExpansionFail rule [] []
+        OIBounded _ (Just ub) | ub < 0 -> ListExpansionFail rule [] []
         OIBounded lb ub
           | (Valid {}, leftover) <- validateTermInList (t : ts) ct ->
               validate leftover (Occur ct (decrementBounds lb ub) : rs)
           | isWithinBoundsInclusive 0 lb ub ->
               validate (t : ts) rs
-          | otherwise -> trace "foo" $ UnapplicableRule "validateList" r
+          | otherwise -> UnapplicableRule "validateList" r
       _ -> case validateTermInList (t : ts) (resolveIfRef cddl r) of
         (Valid {}, leftover) -> validate leftover rs
         (err, _) -> err
@@ -708,7 +712,7 @@ validateList cddl terms rule =
 --------------------------------------------------------------------------------
 -- Maps
 
-validateMap :: CDDL -> [(Term, Term)] -> Rule -> CDDLResult
+validateMap :: HasCallStack => CDDL -> [(Term, Term)] -> Rule -> CDDLResult
 validateMap cddl terms rule =
   case resolveIfRef cddl rule of
     Postlude PTAny -> Valid rule
@@ -718,37 +722,44 @@ validateMap cddl terms rule =
   where
     validate :: [Rule] -> [(Term, Term)] -> [Rule] -> CDDLResult
     validate [] [] [] = Valid rule
-    validate exhausted _ [] = trace (unlines $ show <$> exhausted) $ MapExpansionFail rule [] []
+    validate _ _ [] = MapExpansionFail rule [] []
     validate [] [] (r : rs)
       | isOptional r = validate [] [] rs
       | otherwise = UnapplicableRule "validateMap" r
-    validate exhausted ((k, v) : ts) (r : rs) = case r of
+    validate exhausted kvs (r : rs) = case r of
       Occur ct oi -> case oi of
         OIOptional
-          | (Valid {}, leftover) <- validateKVInMap ((k, v) : ts) ct
+          | (Valid {}, leftover) <- validateKVInMap kvs ct
           , res@Valid {} <- validate [] leftover (exhausted <> rs) ->
               res
-          | otherwise -> validate (r : exhausted) ((k, v) : ts) rs
+          | otherwise -> validate (r : exhausted) kvs rs
         OIZeroOrMore
-          | (Valid {}, leftover) <- validateKVInMap ((k, v) : ts) ct
+          | (Valid {}, leftover) <- validateKVInMap kvs ct
           , res@Valid {} <- validate [] leftover (r : exhausted <> rs) ->
               res
-          | otherwise -> validate (r : exhausted) ((k, v) : ts) rs
-        OIOneOrMore -> case validateKVInMap ((k, v) : ts) ct of
-          (Valid {}, leftover) -> validate [] leftover (Occur ct OIZeroOrMore : exhausted <> rs)
-          (err, _) -> err
-        OIBounded _ (Just ub) | 0 > ub -> traceShow ub $ MapExpansionFail rule [] []
-        OIBounded lb ub
-          | (Valid {}, leftover) <- validateKVInMap ((k, v) : ts) ct
-          , not (isWithinBoundsInclusive 0 lb ub) ->
-              validate [] leftover (Occur ct (decrementBounds lb ub) : exhausted <> rs)
-          | isWithinBoundsInclusive 0 lb ub && not (isValidTerm ((k, v) : ts) ct) ->
-              validate (r : exhausted) ((k, v) : ts) rs
-          | otherwise -> UnapplicableRule "validateMap" r
-      _ -> case validateKVInMap ((k, v) : ts) r of
+          | otherwise -> validate (r : exhausted) kvs rs
+        OIOneOrMore
+          | (Valid {}, leftover) <- validateKVInMap kvs ct
+          , res@Valid {} <- validate [] leftover (Occur ct OIZeroOrMore : exhausted <> rs) ->
+              res
+          | otherwise -> validate (r : exhausted) kvs rs
+        OIBounded mlb mub
+          | Just lb <- mlb, Just ub <- mub, lb > ub -> error "Unsatisfiable range encountered"
+          | otherwise -> case compare 0 <$> mub of
+              Just EQ -> validate exhausted kvs rs
+              Just GT -> error "Unsatisfiable range encountered"
+              _
+                | (Valid {}, leftover) <- validateKVInMap kvs ct
+                , res@Valid {} <-
+                    validate
+                      []
+                      leftover
+                      (Occur ct (decrementBounds mlb mub) : exhausted <> rs) ->
+                    res
+                | otherwise -> validate (r : exhausted) kvs rs
+      _ -> case validateKVInMap kvs r of
         (Valid {}, leftover) -> validate [] leftover (exhausted <> rs)
-        (err, _) -> err
-    validate _ _ _ = error "Impossible happened"
+        _ -> validate (r : exhausted) kvs rs
 
     validateKVInMap ((tk, tv) : ts) (KV k v _) = case (validateTerm cddl tk k, validateTerm cddl tv v) of
       (CBORTermResult _ Valid {}, CBORTermResult _ x@Valid {}) -> (x, ts)
@@ -756,9 +767,6 @@ validateMap cddl terms rule =
       (CBORTermResult _ err, _) -> (err, ts)
     validateKVInMap [] _ = error "No remaining KV pairs"
     validateKVInMap _ x = error $ "Unexpected value in map: " <> show x
-    isValidTerm ts r = case validateKVInMap ts r of
-      (Valid {}, _) -> True
-      _ -> False
 
 --------------------------------------------------------------------------------
 -- Choices
